package teralg;

import java.util.Vector;

/**
 * Elsőbbségi sor adatszerkezet
 * @author Máté Cserép <mcserep@outlook.com>
 */
// Komparátor objektummal tovább általánosítható az összehasonlítás
public class MyPriorityQueue<T extends Comparable> {
    // Kupac reprezentáció
    private Vector<T> heap;
    
    // Új, üres elsőbbségi sor létrehozása
    public MyPriorityQueue() {
        heap = new Vector<T>();
    }
    
    // Sorba
    public void push(T value) {
        // Az új elemet a kupac aljára szúrjuk ...
        heap.add(value);
        // majd felszivárogtatjuk
        liftUp(heap.size() - 1);
    }
    
    // Első
    public T first() {
        if(heap.size() == 0)
            throw new IllegalStateException("The queue is empty.");
        return heap.firstElement();
    }
    
    // Sorból
    public T pop() {
        if(heap.size() == 0)
            throw new IllegalStateException("The queue is empty.");
        
        T result = heap.firstElement();
        // Az alsó, jobb szélső elemet a kupac gyökerébe helyezzük ...
        heap.set(0, heap.lastElement());
        heap.remove(heap.size() - 1);
        // majd lesüllyesszük.
        liftDown(0);
        return result;
    }
    
    // Üres-e
    public boolean empty() {
        return heap.size() == 0;
    }
    
    // Felszivárogtatás
    private void liftUp(int index) {
        int parent = (index - 1) / 2;
        while(index != 0 && heap.get(parent).compareTo(heap.get(index)) < 0)
        {
            // parent és index elem cseréje
            T temp = heap.get(parent);
            heap.set(parent, heap.get(index));
            heap.set(index, temp);
            
            index = parent;
            parent = (index - 1) / 2;
        }
    }
    
    // Lesüllyesztés
    private void liftDown(int index) {
        int left = 2 * index + 1;
        int right = 2 * index + 2;
        while(right < heap.size() && heap.get(index).compareTo(heap.get(right)) < 0 || 
              left < heap.size() && heap.get(index).compareTo(heap.get(left)) < 0) 
        {
            if(right < heap.size() && heap.get(left).compareTo(heap.get(right)) < 0)
            {
                // right és index elem cseréje
                T temp = heap.get(right);
                heap.set(right, heap.get(index));
                heap.set(index, temp);
            
                index = right;
            }
            else
            {
                // left és index elem cseréje
                T temp = heap.get(left);
                heap.set(left, heap.get(index));
                heap.set(index, temp);
            
                index = left;
            }
            left = 2 * index + 1;
            right = 2 * index + 2;
        }
    }
}
